export class Diagram {
		
	private container: any;
	private canvas: HTMLCanvasElement | undefined;
	private ctx: CanvasRenderingContext2D | undefined;
	private initialised = false;
	private unit = 10;
	private xCount = 0;
	private yCount = 0;
	private dynLines = [];
	
	constructor(container: any) {
		this.container = container;
	}
	
	show(diagramData?: any, animation?: boolean): void {
		this.initPanel();
		
		this.draw(diagramData, animation ? true : false);
	}
	
	private initPanel(): void{
		const w = this.container.clientWidth;
		const h = this.container.clientHeight;
				
		const canvas = document.createElement('canvas');		
		var ratio = 1; //window.devicePixelRatio || 1;
		canvas.width = w * ratio; // 实际渲染像素
		canvas.height = h * ratio; // 实际渲染像素
		canvas.style.width = w + 'px'; // 控制显示大小
		canvas.style.height = h + 'px'; // 控制显示大小
		canvas.style.pointerEvents = 'none';
		// canvas.style.position = 'absolute';
		// canvas.style.left = '0';
		// canvas.style.top = '0';
								
		this.container.appendChild(canvas);
				
		this.canvas = canvas;
				
		const ctx = canvas.getContext('2d');
		// ctx.setTransform(ratio, 0, 0, ratio, 0, 0);
		if (ctx) {
			ctx.translate(0.5, 0.5);
			this.ctx = ctx;
		} else {
			throw new Error('Canvas 2D context not found, please check it is running in Browser environment.');
		}

		ctx.lineWidth = 1;
		for(var x = this.unit; x < this.canvas.width; x += this.unit){
			this.xCount++;

			if(this.xCount % 5 == 0)
				ctx.strokeStyle = "#f1f1f1";
			else
				ctx.strokeStyle = "#f7f7f7";
								
			ctx.beginPath();
			ctx.moveTo(x, 0);
			ctx.lineTo(x, this.canvas.height);
			ctx.stroke();
		}
		for(var y = this.unit; y < this.canvas.height; y += this.unit){
			this.yCount++;
			
			if(this.yCount % 5 == 0)
				ctx.strokeStyle = "#f4f4f4";
			else
				ctx.strokeStyle = "#fafafa";
				
			ctx.beginPath();
			ctx.moveTo(0, y);
			ctx.lineTo(this.canvas.width, y);
			ctx.stroke();
		}
		
		this.initialised = true;
	}
		
	private draw(diagramData: any, animation: boolean):void {
		for(var item of diagramData){
			const dash = item.lineStyle == 1 ? [3, 2] : [];
			const color = item.color ? item.color : "#000000";
			const bg = item.background ? item.background : "#ffffff";
			const lineWidth = item.lineWidth > 0 ? item.lineWidth : 1;
			
			this.ctx.setLineDash(dash);
			this.ctx.strokeStyle = color;
			this.ctx.fillStyle = bg;
			this.ctx.lineWidth = lineWidth;
			
			if(item.label){
				this.drawText(item.label);
			} else {
				this.ctx.beginPath();
				if(item.rect){
					const rect = item.rect;								
					this.ctx.rect(rect.x * this.unit, rect.y * this.unit, rect.width * this.unit, rect.height * this.unit);					
				} else if(item.polygon) {
					const path = item.polygon.path;
					for(var i=0;i<path.length;i++){
						if(i == 0)
							this.ctx.moveTo(path[i].x * this.unit, path[i].y * this.unit);
						else
							this.ctx.lineTo(path[i].x * this.unit, path[i].y * this.unit);
					}
					this.ctx.closePath();
				} else if(item.line) {
					const line = item.line;
					if(animation)
						this.drawLine(line[0].x * this.unit, line[0].y * this.unit, line[1].x * this.unit, line[1].y * this.unit, item);
					else{
						this.ctx.moveTo(line[0].x * this.unit, line[0].y * this.unit);
						this.ctx.lineTo(line[1].x * this.unit, line[1].y * this.unit);	
					}
				} else if(item.arc) {
					const arc = item.arc;
					this.ctx.arc(arc.centre.x * this.unit, arc.centre.y * this.unit, arc.r * this.unit, arc.startAngle, arc.endAngle);
				}
				if(!item.fill)
					this.ctx.stroke();
				else
					this.ctx.fill();
			}
		}
		this.dynDrawLine(0);
	}
	
	private drawText(label: any):void{
		const fontSize = 10;
		this.ctx.font = "lighter " + fontSize + "px Microsoft YaHei";
		this.ctx.textBaseline = "bottom";
		
		var rows = [];
		var tmp = "";
		for(var i = 0; i < label.text.length; i++) {
			tmp += label.text[i];
			if(this.ctx.measureText(tmp).width >= label.size.x * this.unit) {
				rows.push(tmp);
				tmp = "";
			}
		}
		if(tmp.length > 0)
			rows.push(tmp);
		
		for(var i = 0; i < rows.length; i++) {
			this.ctx.strokeText(rows[i], label.point.x * this.unit + 2, label.point.y * this.unit + 2 + (i * fontSize));
		}
	}
	
	private drawLine(x1, y1, x2, y2, dataItem):void{
		this.dynLines.push([x1, y1, x2, y2, dataItem]);
	}
	
	private dynDrawLine(idx):void{
		if(idx < this.dynLines.length) {
			this.ctx.beginPath();
			const line = this.dynLines[idx];
			this.dynPaintLine(line[0], line[1], line[2], line[3], line[4], idx);	
		}
	}
	
	private dynPaintLine(x1, y1, x2, y2, dataItem, idx):void{	
		var step = 5;
		const time = 20;
		
		this.ctx.setLineDash(dataItem.lineStyle == 1 ? [3, 2] : []);
		this.ctx.moveTo(x1, y1);
		var next;
		const h = y1 == y2;
		if(h){
			var i = (x2 - x1);
			if(i == 0) {
				this.dynDrawLine(idx + 1);
				return;	
			}
				
			if(i < 0)
				step *= -1;
			if(Math.abs(i) < Math.abs(step))
				next = x1 + i;
			else
				next = x1 + step;
			
			this.ctx.lineTo(next, y1);
			this.ctx.stroke();
			setTimeout(()=>{this.dynPaintLine(next, y1, x2, y2, dataItem, idx);}, time);
		} else {
			var i = (y2 - y1);
			if(i == 0) {
				this.dynDrawLine(idx + 1);
				return;	
			}
				
			if(i < 0)
				step *= -1;
			if(Math.abs(i) < Math.abs(step))
				next = y1 + i;
			else
				next = y1 + step;
				
			this.ctx.lineTo(x1, next);
			this.ctx.stroke();
			setTimeout(()=>{this.dynPaintLine(x1, next, x2, y2, dataItem, idx);}, time);
		}
	}

}

class Point {
	
	private xIdx: number;
	private yIdx: number;
	
	constructor(xIdx: number, yIdx: number) {
		this.xIdx = xIdx;
		this.yIdx = yIdx;
	}
	
	toCanvas(unit: number): Point {
		return new Point(this.xIdx * unit, this.yIdx * unit); 
	}
	
}